import java.io.IOException;
import java.io.Writer;
import java.util.Scanner;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import javax.script.ScriptException;

import org.python.jsr223.PyScriptEngine;
import org.python.jsr223.PyScriptEngineFactory;

public class PythonEngineTest {

    private final AtomicInteger closeCount = new AtomicInteger(0);

    private final AtomicInteger wrongOutputCount = new AtomicInteger(0);

    public static void main(String[] args) {
        PythonEngineTest test = new PythonEngineTest();
        try {
            test.testStdoutDuringParallelExecution(false); // if set to true, the test succeeds
        } catch (InterruptedException e) {
            throw new RuntimeException("Script evaluation didn't terminate in time");
        }
    }

    private void testStdoutDuringParallelExecution(final boolean isSynchronized) throws InterruptedException {

        final int scriptEvalCount = 40;

        final CountDownLatch iterationFinishedLatch = new CountDownLatch(scriptEvalCount);

        for (int i = 0; i < scriptEvalCount; i++) {

            final String index = String.valueOf(i);
            Executors.newCachedThreadPool().execute(new Runnable() {

                @Override
                public void run() {
                    if (isSynchronized) {
                        synchronized (PythonEngineTest.this) {
                            executePythonScript(index, iterationFinishedLatch);
                        }
                    } else {
                        executePythonScript(index, iterationFinishedLatch);
                    }
                }
            });
        }

        assertTrue(iterationFinishedLatch.await(20, TimeUnit.SECONDS));
        assertEquals(scriptEvalCount, closeCount.get());
        assertEquals(0, wrongOutputCount.get());
    }
    
    private void executePythonScript(String index, CountDownLatch iterationFinishedLatch) {
        PyScriptEngineFactory factory = new PyScriptEngineFactory();
        PyScriptEngine engine = (PyScriptEngine) factory.getScriptEngine();
        try {
            Writer outWriter = new WriterStub(index);
            engine.getContext().setWriter(outWriter);
            engine.eval(String.format("import sys\nfor i in range(0, 15):\n print '%s'", index));
            outWriter.close();

            iterationFinishedLatch.countDown();

        } catch (ScriptException | IOException e) {
            throw new RuntimeException(e);
        }
    }

    class WriterStub extends Writer {

        private final StringBuilder builder = new StringBuilder();
        
        private final String index;

        WriterStub(String index) {
            this.index = index;
        }

        @Override
        public void close() throws IOException {
            System.out.println("----\nI'm the Writer instance expecting: " + index + "\nand got:\n" + builder.toString() + "\n");
            try (Scanner scanner = new Scanner(builder.toString())) {
                while (scanner.hasNextLine()) {
                    String line = scanner.nextLine();
                    if (!line.isEmpty() && !line.equals(index)) {
                        System.out.println(String.format("Expected: %s but was: %s", index, line));
                        wrongOutputCount.incrementAndGet();
                    }
                }
            }
            closeCount.incrementAndGet();
        }

        @Override
        public void write(char[] cbuf, int off, int len) throws IOException {
            builder.append(cbuf, off, len);
        }

        @Override
        public void flush() throws IOException {}
    }
    
    private void assertTrue(boolean actual) {
        if (!actual) {
            throw new RuntimeException("Expected: true, but was: " + actual);
        }
    }
    
    private void assertEquals(int expected, int actual) {
        if (expected != actual) {
            throw new RuntimeException(String.format("Expected %d but was: %d", expected, actual));
        }
    }
}
